from mailer import Mailer, Message

from .. import root, config, logger

LOGIN_HTML = root / 'html' / 'login.html'


class EmailMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def send_auth_email(self, email, auth_url):
        logger.info(f'Login here to use Spotify API: {auth_url}')

        mailer = Mailer(**config.email)
        html_content = LOGIN_HTML.read_text().replace('SPOTIFY_AUTHENTICATION_URL', auth_url)

        message = Message(
            From=config.email.usr or email,
            To=email,
            charset="utf-8")
        message.Subject = "Spotify API Authentication"
        message.Html = html_content
        message.Body = f'Login here to use Spotify API: {auth_url}'

        mailer.send(message)



import time
import random
from datetime import date, timedelta
from itertools import chain

from pyorderby import orderby

from .. import logger
from ..cache import *
from ..constants import TimeRange, AudioFeature


class RecommenderMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @db_session
    def fetch_playlists(self):
        fetched_ids = set(select(p.id for p in Playlist))
        results = chain(
            self.user_playlists('particledetector').all(),
            self.user_playlists('thesoundsofspotify').all()
        )

        for playlist in results:
            logger.info(f'Fetching {playlist.name}')
            if playlist.id not in fetched_ids:
                Playlist.from_dict(playlist)
            fetched_ids.add(playlist.id)

    @db_session
    def fetch_user_top(self, time_range):
        self.user.top_artists.clear()
        disliked_artists = set(self.user.disliked_artists.id.distinct().keys())
        for artist in self.current_user_top_artists(limit=50, time_range=time_range).iterall():
            if artist.id in disliked_artists:
                continue

            if Artist.exists(id=artist.id):
                self.user.top_artists.add(Artist[artist.id])
            else:
                self.user.top_artists.add(Artist.from_dict(artist))

        self.user.top_genres = self.user.top_artists.genres.distinct().keys() - self.user.disliked_genres

        if self.user.top_expires_at is None:
            self.user.top_expires_at = {}
        self.user.top_expires_at[TimeRange(time_range).value] = time.mktime((date.today() + timedelta(days=1)).timetuple())

    @db_session
    def genre_playlist(self, genre, popularity=Playlist.Popularity.SOUND):
        return Playlist.get(genre=genre, popularity=Playlist.Popularity(popularity).value)

    @db_session
    def top_artists(self, time_range=TimeRange.SHORT_TERM):
        if self.user.top_expired(time_range):
            self.fetch_user_top(time_range)

        return self.user.top_artists

    @db_session
    def top_genres(self, time_range=TimeRange.SHORT_TERM):
        if self.user.top_expired(time_range):
            self.fetch_user_top(time_range)

        return self.user.top_genres

    def order_by(self, features, tracks):
        audio_features = self.audio_features(tracks=tracks)
        if isinstance(features, AudioFeature):
            features = features.value

        return sorted(audio_features, key=orderby(features))

    def fill_with_related_artists(self, artists, limit=5):
        if len(artists) >= limit:
            return artists

        artist_set = set(artists)
        artist_list = list(artists)
        disliked_artists = set(self.user.disliked_artists.id.distinct().keys())

        while len(artist_set) < limit:
            related_artists = {
                a.id for a in self.artist_related_artists(random.choice(artist_list))
                if a.id not in disliked_artists}

            related_artists_limit = min(random.randint(1, limit - len(artist_set)), len(related_artists))
            artist_set |= set(random.sample(related_artists, related_artists_limit))

        return artist_set

    @db_session
    def recommend_by_top_artists(self, artist_limit=2, track_limit=100, use_related=True, features_order=None, time_range=TimeRange.SHORT_TERM, **kwargs):
        """Get a list of recommended songs.

        Returns:
            list: List of tracks
        """

        artists = self.top_artists(time_range=time_range).select().without_distinct().random(artist_limit)
        artists = self.fill_with_related_artists([a.id for a in artists])

        tracks = self.recommendations(seed_artists=artists, limit=track_limit, **kwargs)

        if features_order:
            tracks = self.order_by(features_order, tracks)

        return tracks

    @db_session
    def dislike_artist(self, artist):
        if isinstance(artist, str):
            artist = get(a for a in Artist if a.id == artist or a.name == artist)
        elif isinstance(artist, dict):
            artist = Artist.get(artist.id) or Artist.from_dict(artist)

        self.user.dislike(artist=artist)

    @db_session
    def dislike_genre(self, genre):
        if isinstance(genre, str):
            genre = genre.lower()
            genre = Genre.get(name=genre) or Genre(name=genre)

        self.user.dislike(genre=genre)



import uuid
import socket
import threading
from pathlib import Path

import hug
from cachecontrol import CacheControlAdapter
from oauthlib.oauth2 import BackendApplicationClient
from requests_oauthlib import OAuth2Session
from wsgiref.simple_server import make_server
from cachecontrol.caches.file_cache import FileCache

from .. import root, config, logger
from ..cache import User, get, select, db_session
from ..constants import API, AuthFlow, AllScopes
from ..exceptions import SpotifyCredentialsException

AUTH_HTML_FILE = root / 'html' / 'auth_message.html'
CACHE_FILE = Path.home() / '.cache' / 'spfy' / '.web_cache'


class AuthMixin:
    def __init__(self, client_id=None, client_secret=None, redirect_uri=None, userid=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client_id = client_id or config.app.client_id
        self.client_secret = client_secret or config.app.client_secret
        self.redirect_uri = self._get_redirect_uri(redirect_uri)

        self.session = None
        self.userid = userid

        self.callback_reached = threading.Event()

    def _get_redirect_uri(self, redirect_uri):
        redirect_uri = (
            redirect_uri or
            config.app.redirect_uri or
            f'http://{socket.gethostname()}.local')

        if config.auth.callback.enabled and config.auth.callback.port and redirect_uri:
            redirect_uri += f':{config.auth.callback.port}'

        return redirect_uri

    def get_session(self, *args, **kwargs):
        session = OAuth2Session(*args, **kwargs)
        cache_adapter = CacheControlAdapter(
            cache=FileCache(CACHE_FILE),
            pool_connections=config.http.connections,
            pool_maxsize=config.http.connections,
            max_retries=config.http.retries)
        session.mount('http://', cache_adapter)
        return session

    @property
    @db_session
    def user(self):
        return User[self.userid]

    @property
    def is_authenticated(self):
        return bool(self.session and self.session.token)

    @db_session
    def authenticate_user(self, username=None, email=None, code=None, state=None, auth_response=None, scope=AllScopes):
        session = self.session or self.get_session(self.client_id, redirect_uri=self.redirect_uri, scope=scope, auto_refresh_url=API.TOKEN.value)

        if self.userid:
            user = User.get(id=self.userid)
            if user:
                session.token = user.token
                session.token_updater = User.token_updater(user.id)
                return session

        if username or email:
            user = get(u for u in User if u.username == username or u.email == email)
            if user:
                self.userid = user.id
                session.token = user.token
                session.token_updater = User.token_updater(user.id)
                return session

        if code or auth_response:
            token = session.fetch_token(
                token_url=API.TOKEN.value,
                client_id=self.client_id,
                client_secret=self.client_secret,
                code=code, state=state,
                authorization_response=auth_response)

            user_details = self.current_user()
            user = select(u for u in User if u.username == user_details.id or u.email == user_details.email).for_update().get()

            if user:
                user.token = token
                if user.id != self.userid:
                    self.userid = user.id
            elif self.userid:
                user = User.get_for_update(id=self.userid)
                if user:
                    user.token = token

            if not user:
                self.userid = self.userid or uuid.uuid4()
                user = User(id=self.userid, username=user_details.id, email=user_details.email, token=token)

            return session

        return session

    @db_session
    def authenticate_server(self):
        default_user = User.default()
        self.userid = default_user.id

        session = self.session or self.get_session(client=BackendApplicationClient(self.client_id))
        session.token_updater = User.token_updater(default_user.id)

        if default_user.token:
            session.token = default_user.token
        else:
            default_user.token = session.fetch_token(
                token_url=API.TOKEN.value,
                client_id=self.client_id,
                client_secret=self.client_secret)

        return session

    def authenticate(self, flow=config.auth.flow, **auth_params):
        if not (self.client_id and self.client_secret):
            raise SpotifyCredentialsException

        flow = AuthFlow(flow)
        if flow == AuthFlow.CLIENT_CREDENTIALS:
            self.session = self.authenticate_server()
        elif flow == AuthFlow.AUTHORIZATION_CODE:
            self.session = self.authenticate_user(**auth_params)
            if not self.session.token:
                if config.auth.callback.enabled:
                    self.start_callback()

                authorization_url, _ = self.session.authorization_url(API.AUTHORIZE.value)
                if config.auth.send_auth_url_to_email:
                    email = auth_params.get('email') or config.auth.email
                    self.send_auth_email(email, authorization_url)
                else:
                    print(f'Login here: {authorization_url}')

                self.wait_for_authorization()

    def wait_for_authorization(self):
        if not config.auth.callback.enabled:
            url = input('Paste the URL you are redirected to:')
            self.session = self.authenticate_user(auth_response=url)
        else:
            self.callback_reached.wait(config.auth.callback.timeout)
            self.stop_callback()

    def stop_callback(self):
        if self.httpd:
            self.httpd.shutdown()

    def start_callback(self):
        self.callback_reached.clear()

        @hug.get('/', output=hug.output_format.html)
        def callback(code: hug.types.text, state: hug.types.text):
            html = AUTH_HTML_FILE.read_text()
            try:
                self.session = self.authenticate_user(code=code, state=state)
                html = html.replace('SPOTIFY_AUTH_MESSAGE', 'Successfully logged in!')
                html = html.replace('BACKGROUND_COLOR', '#65D46E')
            except Exception as exc:
                logger.exception(exc)
                html = html.replace('SPOTIFY_AUTH_MESSAGE', 'Could not get authorization token.')
                html = html.replace('BACKGROUND_COLOR', '#EC2E50')
            finally:
                self.callback_reached.set()

            return html

        api = __hug__.http.server(None)  # noqa
        self.httpd = make_server('', config.auth.callback.port, api)

        threading.Thread(target=self.httpd.serve_forever, daemon=True).start()



import os
import threading
from collections import OrderedDict

from first import first
from cached_property import cached_property

from .. import config
from ..volume import (
    AlsaVolumeControl,
    LinuxVolumeControl,
    SpotifyVolumeControl,
    ApplescriptVolumeControl
)
from ..constants import VolumeBackend


class PlayerMixin:
    def __init__(self, device=None, alsa_device=None, alsa_mixer=None, speaker=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.device = device or config.player.device
        self.alsa_device = alsa_device or config.player.alsa.device
        self.alsa_mixer = alsa_mixer or config.player.alsa.mixer
        self.speaker = speaker or config.player.speaker

    @cached_property
    def _optimal_backend(self):
        return first(self._backends.values())

    @cached_property
    def _backends(self):
        return OrderedDict({
            VolumeBackend.APPLESCRIPT: self._applescript_volume_control,
            VolumeBackend.LINUX: self._linux_volume_control,
            VolumeBackend.ALSA: self._alsa_volume_control,
            VolumeBackend.SPOTIFY: self._spotify_volume_control
        })

    @cached_property
    def _applescript_volume_control(self):
        try:
            if os.uname().sysname != 'Darwin':
                return
        except:
            return

        return ApplescriptVolumeControl(self.speaker)

    @cached_property
    def _linux_volume_control(self):
        try:
            if os.uname().sysname != 'Linux':
                return
        except:
            return

        return LinuxVolumeControl(self, self.alsa_mixer, spotify_device=self.device, alsa_device=self.alsa_device)

    @cached_property
    def _alsa_volume_control(self):
        try:
            if os.uname().sysname != 'Linux':
                return
        except:
            return

        return AlsaVolumeControl(self.alsa_mixer, device=self.alsa_device)

    @cached_property
    def _spotify_volume_control(self):
        return SpotifyVolumeControl(self, device=self.device)

    def backend(self, backend=None):
        if not backend:
            return self._optimal_backend

        volume_backend = self._backends[VolumeBackend(backend)]
        if not volume_backend:
            raise ValueError(f'Backend {volume_backend} is not available on this system')

        return volume_backend

    def change_volume(self, value=0, backend=None):
        volume_backend = self.backend(backend)
        volume = volume_backend.volume + value
        volume_backend.volume = volume
        return volume

    def volume_up(self, backend=None):
        return self.change_volume(value=+1, backend=backend)

    def volume_down(self, backend=None):
        return self.change_volume(value=-1, backend=backend)

    def fade_up(self, **kwargs):
        self.fade(**{**config.volume.fade.up, **kwargs})

    def fade_down(self, **kwargs):
        self.fade(**{**config.volume.fade.down, **kwargs})

    def fade(self, limit, start, step, seconds, force=False, backend=None, spotify_volume=100):
        volume_backend = self.backend(backend)
        if not isinstance(volume_backend, SpotifyVolumeControl):
            self.change_volume(spotify_volume, VolumeBackend.SPOTIFY)

        kwargs = dict(limit=limit, start=start, step=step, seconds=seconds, force=force)
        threading.Thread(target=volume_backend.fade, kwargs=kwargs).start()



import os
import re
import time
from enum import IntEnum
from uuid import UUID, NAMESPACE_URL, uuid4, uuid5
from datetime import date, datetime

from pony.orm import *

from .. import config
from ..constants import TimeRange

if os.getenv('DEBUG'):
    sql_debug(True)
    import logging
    logging.getLogger('pony.orm.sql').setLevel(logging.DEBUG)

db = Database()


class User(db.Entity):
    DEFAULT_EMAIL = 'spfy@backend'
    DEFAULT_USERNAME = 'spfy-backend'
    DEFAULT_USERID = uuid5(NAMESPACE_URL, DEFAULT_USERNAME)

    id = PrimaryKey(UUID, default=uuid4)
    email = Required(str, unique=True, index=True)
    username = Required(str, unique=True, index=True)
    token = Required(Json)
    api_calls = Required(int, default=0, volatile=True)
    created_at = Required(datetime, default=datetime.now)
    last_usage_at = Required(datetime, default=datetime.now)

    top_artists = Set('Artist')
    disliked_artists = Set('Artist')

    top_genres = Set('Genre')
    disliked_genres = Set('Genre')

    top_expires_at = Optional(Json)

    def dislike(self, artist=None, genre=None):
        assert artist or genre
        if artist:
            self.disliked_artists.add(artist)
            self.top_artists.remove(artist)
        if genre:
            self.disliked_genres.add(genre)
            self.top_genres.remove(genre)

    def top_expired(self, time_range):
        time_range = TimeRange(time_range).value
        return (
            not self.top_expires_at or
            time_range not in self.top_expires_at or
            time.time() >= self.top_expires_at[time_range])

    @classmethod
    def default(cls):
        try:
            return User[cls.DEFAULT_USERID]
        except ObjectNotFound:
            return User(id=cls.DEFAULT_USERID, username=cls.DEFAULT_USERNAME, email=cls.DEFAULT_EMAIL, token={})

    @staticmethod
    def token_updater(id):
        @db_session
        def update(token):
            if isinstance(token, ormtypes.TrackedDict):
                token = token.get_untracked()
            User[id].token = token

        return update


class Image(db.Entity):
    url = PrimaryKey(str)
    height = Optional(int)
    width = Optional(int)
    playlist = Optional('Playlist')
    artist = Optional('Artist')


class SpotifyUser(db.Entity):
    id = PrimaryKey(str)
    name = Optional(str)
    playlists = Set('Playlist')

    @property
    def uri(self):
        return f'spotify:user:{self.id}'

    @property
    def href(self):
        return f'https://api.spotify.com/v1/users/{self.id}'

    @property
    def external_url(self):
        return f'http://open.spotify.com/user/{self.id}'


class Genre(db.Entity):
    name = PrimaryKey(str)
    artists = Set('Artist')
    playlists = Set('Playlist')
    fans = Set(User, reverse='top_genres', table='genre_fans')
    haters = Set(User, reverse='disliked_genres', table='genre_haters')


class Playlist(db.Entity):
    PARTICLE_RE = re.compile('The (?P<popularity>Sound|Pulse|Edge) of (?P<genre>.+)')
    SOUND_CITY_RE = re.compile('The Sound of (?P<city>.+) (?P<country_code>[A-Z]{2})')
    NEEDLE_RE = re.compile('The Needle / (?P<country>.+) (?P<date>[0-9]{8}) - (?P<popularity>Current|Emerging|Underground)')
    PINE_NEEDLE_RE = re.compile('The Pine Needle / (?P<country>.+)')

    class Popularity(IntEnum):
        SOUND = 0
        PULSE = 1
        EDGE = 2

        CURRENT = 3
        EMERGING = 4
        UNDERGROUND = 5

    id = PrimaryKey(str)
    collaborative = Required(bool)
    images = Set(Image, cascade_delete=True)
    name = Required(str)
    owner = Required(SpotifyUser)
    public = Required(bool)
    snapshot_id = Required(str)
    tracks = Required(int)
    popularity = Optional(int, index=True)
    genre = Optional(Genre)
    country = Optional(str, index=True)
    country_code = Optional(str, index=True, max_len=2)
    city = Optional(str, index=True)
    date = Optional(date)
    christmas = Optional(bool, index=True)
    composite_key(genre, popularity)

    @property
    def uri(self):
        return f'spotify:user:{self.owner.id}:playlist:{self.id}'

    @property
    def href(self):
        return f'https://api.spotify.com/v1/users/{self.owner.id}/playlists/{self.id}'

    @property
    def external_url(self):
        return f'http://open.spotify.com/user/{self.owner.id}/playlist/{self.id}'

    @classmethod
    def from_dict(cls, playlist):
        owner = (
            SpotifyUser.get(id=playlist.owner.id) or
            SpotifyUser(id=playlist.owner.id, name=playlist.owner.get('display_name', playlist.owner.id))
        )
        fields = {
            'id': playlist.id,
            'collaborative': playlist.collaborative,
            'name': playlist.name,
            'owner': owner,
            'public': playlist.public,
            'snapshot_id': playlist.snapshot_id,
            'tracks': playlist.tracks.total,
            'images': [Image.get(url=im.url) or Image(**im) for im in playlist.images]
        }

        match = cls.SOUND_CITY_RE.match(playlist.name)
        if match:
            city, country_code = match.groups()
            fields['popularity'] = cls.Popularity.SOUND.value
            fields['city'] = city.lower()
            fields['country_code'] = country_code
            return cls(**fields)

        match = cls.PARTICLE_RE.match(playlist.name)
        if match:
            popularity, genre = match.groups()
            genre = genre.lower()
            fields['popularity'] = cls.Popularity[popularity.upper()].value
            fields['genre'] = Genre.get(name=genre) or Genre(name=genre)
            fields['christmas'] = 'christmas' in genre
            return cls(**fields)

        match = cls.PINE_NEEDLE_RE.match(playlist.name)
        if match:
            country = match.groups()[0]
            fields['popularity'] = cls.Popularity.CURRENT.value
            fields['country'] = country.lower()
            fields['christmas'] = True
            return cls(**fields)

        match = cls.NEEDLE_RE.match(playlist.name)
        if match:
            country, date, popularity = match.groups()
            fields['popularity'] = cls.Popularity[popularity.upper()].value
            fields['country'] = country.lower()
            fields['date'] = datetime.strptime(date, '%Y%m%d').date()
            return cls(**fields)

        return cls(**fields)


class Artist(db.Entity):
    id = PrimaryKey(str)
    name = Required(str, index=True)
    followers = Required(int)
    genres = Set(Genre)
    images = Set(Image, cascade_delete=True)
    fans = Set(User, reverse='top_artists', table='artist_fans')
    haters = Set(User, reverse='disliked_artists', table='artist_haters')
    popularity = Optional(int)

    @property
    def uri(self):
        return f'spotify:artist:{self.id}'

    @property
    def href(self):
        return f'https://api.spotify.com/v1/artists/{self.id}'

    @property
    def external_url(self):
        return f'http://open.spotify.com/artist/{self.id}'

    @classmethod
    def from_dict(cls, artist):
        genres = [Genre.get(name=genre) or Genre(name=genre) for genre in artist.genres]
        images = [Image.get(url=image.url) or Image(**image) for image in artist.images]
        return cls(
            id=artist.id, name=artist.name, followers=artist.followers.total,
            genres=genres, images=images, popularity=artist.popularity)


if config.database.filename:
    config.database.filename = os.path.expandvars(config.database.filename)

db.bind(**config.database)
db.generate_mapping(create_tables=True)



__version__ = '2.6.6'

import pathlib  # isort:skip
root = pathlib.Path(__file__).parent  # isort:skip

APP_NAME = 'SPFY'

import os  # isort:skip
ENV = os.getenv(f'{APP_NAME.upper()}_ENV', 'config')  # isort:skip

import kick  # isort:skip
kick.start(f'{APP_NAME.lower()}', config_variant=ENV)  # isort:skip

from kick import config, logger  # isort:skip

from .client import SpotifyClient
from .result import SpotifyResult
from .wrapper import Spotify
from .constants import *
from .exceptions import *



# -*- coding: utf-8 -*-

import abc
import subprocess
from time import sleep

from . import logger
from .constants import VOLUME_FADE_SECONDS

try:
    import alsaaudio
except:
    pass


class VolumeControl(abc.ABC):
    @abc.abstractmethod
    def unmute(self):
        pass

    @abc.abstractmethod
    def mute(self):
        pass

    @abc.abstractproperty
    def volume(self):
        pass

    @volume.setter
    def volume(self, val):
        pass

    @abc.abstractmethod
    def should_stop_fading(self, device_volume, old_volume):
        return abs(device_volume - old_volume) > 2

    def fade(self, limit=100, start=1, step=1, seconds=VOLUME_FADE_SECONDS, force=False):
        self.unmute()

        delay = seconds / ((limit - start) / step)
        device_volume = self.volume

        self.volume = start

        for next_volume in range(start + step, limit + 1, step):
            sleep(delay)
            device_volume = self.volume
            old_volume = next_volume - step

            if not force and self.should_stop_fading(device_volume, old_volume):
                logger.debug(f'''A stop fading condition was met:
                    Current volume: {device_volume}
                    Old volume: {old_volume}''')
                break

            logger.debug(f'Setting volume to {next_volume}')
            self.volume = next_volume


class SpotifyVolumeControl(VolumeControl):
    def __init__(self, client, device=None):
        self.spotify = client
        self.device = device
        self.old_volume = None

    def mute(self):
        self.old_volume = self.volume
        self.spotify.volume(0, device=self.device)

    def unmute(self):
        if self.old_volume:
            self.volume = self.old_volume
        else:
            self.volume = 1

    def should_stop_fading(self, device_volume, old_volume):
        is_playing = self.spotify.current_playback().is_playing
        logger.debug(f'Spotify playing: {is_playing}')
        return (
            super().should_stop_fading(device_volume, old_volume) or
            not is_playing)

    @property
    def volume(self):
        return int(self.spotify.get_device(device=self.device).volume_percent) + 1

    @volume.setter
    def volume(self, val):
        self.spotify.volume(min(max(val, 1), 100), device=self.device)


class AlsaVolumeControl(VolumeControl):
    def __init__(self, mixer_name, device=None):
        if isinstance(device, str):
            kwargs = dict(device=device)
        elif isinstance(device, int):
            kwargs = dict(cardindex=device)
        else:
            kwargs = dict()

        self.mixer_name = mixer_name
        self.device = device
        self.mixer = alsaaudio.Mixer(mixer_name, **kwargs)

    def mute(self):
        self.mixer.setmute(1)

    def unmute(self):
        self.mixer.setmute(0)

    def should_stop_fading(self, device_volume, old_volume):
        is_mute = self.mixer.getmute()[0]
        logger.debug(f'Mute status: {is_mute}')
        return (
            super().should_stop_fading(device_volume, old_volume) or
            is_mute)

    @property
    def volume(self):
        vol = self.mixer.getvolume()
        return sum(vol) // len(vol)

    @volume.setter
    def volume(self, val):
        self.mixer.setvolume(max(val, 1))


class ApplescriptVolumeControl(VolumeControl):
    def __init__(self, device=None):
        self.device = device
        self.old_volume = None

    def fade(self, limit=100, start=1, step=1, seconds=VOLUME_FADE_SECONDS, force=False):
        if self.device:
            self.switch_audio_device(self.device)

        super().fade(limit, start, step, seconds, force)

    def osascript(self, cmd):
        cmd = f"/usr/bin/osascript -e '{cmd}'"
        result = subprocess.check_output(cmd, shell=True).decode().strip()

        logger.debug('Command: %s', cmd)
        logger.debug('Output: %s', result)

        return result

    def switch_audio_device(self, device):
        logger.debug(f'Switching audio device: {device}')
        subprocess.call(f"/usr/local/bin/SwitchAudioSource -s \"{device}\"", shell=True)

    def should_stop_fading(self, device_volume, old_volume):
        is_playing = self.osascript('tell application "Spotify" to get player state') == 'playing'
        logger.debug(f'Spotify playing: {is_playing}')
        return (
            super().should_stop_fading(device_volume, old_volume) or
            not is_playing)

    def spotify_mute(self):
        self.old_volume = self.spotify_volume
        self.osascript('tell application "Spotify" to set sound volume to 0')

    def spotify_unmute(self):
        if self.old_volume:
            self.spotify_volume = self.old_volume
        else:
            self.spotify_volume = 1

    def mute(self):
        self.osascript('set volume with output muted')

    def unmute(self):
        self.osascript('set volume without output muted')

    @property
    def spotify_volume(self):
        return int(self.osascript('tell application "Spotify" to get sound volume')) + 1

    @spotify_volume.setter
    def spotify_volume(self, val):
        self.osascript(f'tell application "Spotify" to set sound volume to {max(val, 1)}')

    @property
    def system_volume(self):
        return int(self.osascript('output volume of (get volume settings)'))

    @system_volume.setter
    def system_volume(self, val):
        self.osascript(f'set volume output volume {max(val, 1)}')

    @property
    def volume(self):
        return self.system_volume

    @volume.setter
    def volume(self, val):
        self.old_volume = self.volume
        self.system_volume = val
        self.spotify_volume = val // 2


class LinuxVolumeControl(AlsaVolumeControl):
    def __init__(self, spotify_client, alsa_mixer_name, spotify_device=None, alsa_device=None):
        super().__init__(alsa_mixer_name, device=alsa_device)
        self.spotify_volume_control = SpotifyVolumeControl(spotify_client, spotify_device)

    def should_stop_fading(self, device_volume, old_volume):
        return (
            super().should_stop_fading(device_volume, old_volume) or
            self.spotify_volume_control.should_stop_fading(device_volume, old_volume))

    def get_amixer_cmd(self, volume):
        cmd = ['/usr/bin/amixer']
        if self.device:
            flag = '-c' if isinstance(self.device, int) else '-D'
            cmd += [flag, self.device]

        cmd += ['sset', self.mixer_name, 'unmute', f'{volume}%']

        logger.debug(cmd)

        return cmd

    def fade(self, limit=100, start=1, step=1, seconds=VOLUME_FADE_SECONDS, force=False):
        subprocess.call(self.get_amixer_cmd(start))

        super().fade(limit, start, step, seconds, force)



class SpotifyException(Exception):
    def __init__(self, response=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.http_status_code = response.status_code
        self.headers = response.headers or {}

        if response.text and len(response.text) > 0 and response.text != 'null':
            self.msg = f'{response.url}:\n {response.json()["error"]["message"]}'
        else:
            self.msg = f'{response.url}: error'

    def __str__(self):
        return f'''
        HTTP Status Code: {self.http_status_code}
        {self.msg}'''


class SpotifyRateLimitException(SpotifyException):
    def __init__(self, retry_after=0, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.retry_after = retry_after


class SpotifyForbiddenException(SpotifyException):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class SpotifyCredentialsException(Exception):
    def __str__(self):
        return '''
        You need to set your Spotify API credentials in ~/.config/spfy/config.toml

        Get your credentials at
            https://developer.spotify.com/my-applications
        '''


class SpotifyAuthException(Exception):
    def __str__(self):
        return '''
        You need to authenticate before making any request.
        '''



import os
import pathlib

from setuptools import setup, find_packages

CONFIGDIR = pathlib.Path.home() / '.config' / 'spfy'
CONFIGDIR.mkdir(parents=True, exist_ok=True)

with open('spfy/__init__.py', 'r') as f:
    for line in f:
        if line.startswith('__version__'):
            version = line.strip().split('=')[1].strip(' \'"')
            break
    else:
        version = '0.0.1'

with open('README.md', 'rb') as f:
    readme = f.read().decode('utf-8')

REQUIRES = [
    'addict',
    'backoff',
    'fire',
    'first',
    'gunicorn',
    'hug',
    'oauthlib',
    'pyorderby',
    'requests',
    'requests_oauthlib',
    'mailer',
    'pony',
    'cached_property',
    'kick',
    'cachecontrol',
    'lockfile',
]

try:
    if os.uname().sysname == 'Linux':
        REQUIRES.append('pyalsaaudio')
except:
    pass

setup(
    name='spfy',
    version=version,
    description='',
    long_description=readme,
    author='Alin Panaitiu',
    author_email='alin.p32@gmail.com',
    maintainer='Alin Panaitiu',
    maintainer_email='alin.p32@gmail.com',
    url='https://github.com/alin23/spfy',
    license='MIT/Apache-2.0',

    keywords=[
        'spotify', 'spfy'
    ],

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],

    install_requires=REQUIRES,
    tests_require=['coverage', 'pytest'],

    packages=find_packages(),
    package_data={
        'spfy': [
            'config/*.toml',
            'html/*.html',
        ]
    },
    data_files=[
        (str(CONFIGDIR), ['spfy/config/config.toml'])
    ],

    entry_points={
        'console_scripts': ['spotify = spfy.wrapper:main']
    },

)



import threading
from itertools import chain
from urllib.parse import parse_qs, urlparse, urlunparse
from concurrent.futures import ThreadPoolExecutor

import addict
from cached_property import cached_property

from . import config

LOCAL_ATTRIBUTES = {'_client', '_next_result', '_next_result_available'}


class SpotifyResult(addict.Dict):
    ITER_KEYS = ('items', 'artists', 'tracks', 'albums', 'audio_features')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._next_result_available = threading.Event()
        self._next_result = None

    def __iter__(self):
        for key in self.ITER_KEYS:
            if key in self:
                return iter(self[key])
        return super().__iter__()

    def __getitem__(self, item):
        if isinstance(item, int):
            for key in self.ITER_KEYS:
                if key in self:
                    return self[key][item]
        return super().__getitem__(item)

    @classmethod
    def _hook(cls, item):
        if isinstance(item, dict):
            return addict.Dict(item)
        elif isinstance(item, (list, tuple)):
            return type(item)(cls._hook(elem) for elem in item)
        return item

    def items(self):
        return filter(lambda i: i[0] not in LOCAL_ATTRIBUTES, super().items())

    def keys(self):
        return super().keys() - LOCAL_ATTRIBUTES

    def values(self):
        return [v for k, v in self.items()]

    @cached_property
    def base_url(self):
        return urlunparse([*urlparse(self.href)[:3], '', '', ''])

    def _get_with_params(self, params):
        return self._client._get(self.base_url, **params)

    def get_next_params_list(self):
        if self['next'] and self['href']:
            url = urlparse(self['href'])
            params = {k: v[0] for k, v in parse_qs(url.query).items()}
            limit = int(params.pop('limit', 20))
            offset = int(params.pop('offset', 0))
            return [
                {**params, 'limit': 50, 'offset': off}
                for off in range(offset + limit, self.total, 50)]
        else:
            return []

    def all(self):
        params_list = self.get_next_params_list()
        if not params_list:
            return []

        with ThreadPoolExecutor(max_workers=config.http.concurrent_connections) as executor:
            return chain.from_iterable(executor.map(self._get_with_params, params_list, timeout=10 * len(params_list)))

    @cached_property
    def next(self):
        if self['next']:
            return self._client._get(self['next'])
        else:
            return None

    def _fetch_next_result(self, result):
        self._next_result = result.next
        self._next_result_available.set()

    def iterall(self):
        result = self

        while result:
            threading.Thread(target=self._fetch_next_result, args=(result,)).start()
            for item in result:
                yield item

            if self._next_result_available.wait(10):
                result = self._next_result
                self._next_result_available.clear()
            else:
                break
